local ti=table.insert
local mx,mn=math.max,math.min

local function lerp(u,p1,p2)
  return p1+(p2-p1)*u
end

local function invLerp(v,p1,p2)
  return (v-p1)/(p2-p1)
end

local function bezier(u,a,b,c,d)
  local omt=1.0-u;
  local omt2=omt*omt;
  local omt3=omt2*omt;
  local u2=u*u;
  local u3=u2*u;
  return a*omt3 + b*omt2*u*3.0 + c*omt*u2*3.0 + d*u3;
end

function bakeBezierPath(points,interval,step,iterations)
  interval=interval or 5.0
  step=step or 0.1
  iterations=iterations or 10
  
  local pos,dist=vec2(points[1].x,points[1].y),0.0
  local bakedPoint,bakedDist={pos},{dist}
  local cost=0
  
  for i=1,#points-1,3 do
    local p=0.0
    
    while p<1.0 do
      local np=p+step if np>1.0 then np=1.0 end

      local npp=bezier(np,points[i],points[i+1],points[i+2],points[i+3])
      local d=pos:dist(npp)
      cost=cost+1

      if d>interval then -- distance to big, look for a closer npp

        -- binary search for a closer npp
        local low,hi=p,np
        local mid=low+(hi-low)*0.5
        
        for j=1,iterations do
          npp=bezier(mid,points[i],points[i+1],points[i+2],points[i+3])
          d=pos:dist(npp)
          cost=cost+1
          if interval < d then hi=mid else low=mid end
          mid=low+(hi-low)*0.5;
        end
        
        pos=npp
        p=mid
        dist=dist+d
        
        ti(bakedPoint,pos)
        ti(bakedDist,dist)
      else
        p=np
      end
    end
    
    ti(bakedPoint,vec2(points[i+3].x,points[i+3].y))
    ti(bakedDist,dist+pos:dist(points[i+3]))
  end
  
  return bakedPoint,bakedDist,cost
end

local function getPointFromDist(bakedPoint,bakedDist,dist)
  for i=1,#bakedPoint do
    if dist<bakedDist[i] then
      local h=mx(1,i-1)
      local p1,p2=bakedPoint[h],bakedPoint[i]
      local d1,d2=bakedDist[h],bakedDist[i]
      local u=invLerp(dist,d1,d2)
      return lerp(u,p1,p2)
    elseif dist==bakedDist[#bakedDist] then
      return bakedPoint[#bakedPoint]
    end
  end
  return vec2()
end

function getPointOnPathFromDist(bakedPoint,bakedDist,dist)
  dist=mx(0,mn(dist,bakedDist[#bakedDist]))
  return getPointFromDist(bakedPoint,bakedDist,dist)
end

function getPointOnPathFromUnit(bakedPoint,bakedDist,unit)
  unit=mx(0.0,mn(unit,1.0))
  local dist=bakedDist[#bakedDist]*unit
  return getPointFromDist(bakedPoint,bakedDist,dist)
end
